////////////////////////////////////////////////////
// kello_hw4_v092.c
////////////////////////////////////////////////////
// Copyright 2017-2018 Jaakko Kairus
// jaakko@kair.us
//
// You may freely use and modify this software for personal use.
// Commercial use of this software or parts of it is prohibited
// without written permission from author.
// DISCLAIMER: This software comes with ABSOLUTELY NO WARRANTY
//
// For more information about this project see project web page at
// http://kair.us/projects/clock/v4/index.html
//
// Compiled with CCS compiler version 5.081
// Total compiled length 16023 words
///////////////////////////////////////////////////
// Version history
// v0.1 13.2.2017
//      Initial version. Support for TLC5925 and PCF2129. Compiler 5.054. Length 1260 words.
// v0.2 30.6.2017
//		Added initial support for remote (NEC coding). Length 1771 words.
// v0.3 6.9.2017
//		Moved remote decoding stuff to separate file (ir_machine.c) and changed remote
//		data type to a struct compatible with irmp (https://www.mikrocontroller.net/articles/IRMP)
//		Not however using IRMP to avoid getting involved with all the GNU GPL stuff.
//		Added support for showing remote control codes and learn mode to teach clock
//		to use any NEC protocol remote. Learned codes are stored to PIC EEPROM. Added
//		support to set date and time. Added support to adjust brightness. Brightness
//		and display mode are stored to PCF2129 NVRAM (using alarm registers).
//		Added temperature readout for two 10k NTC sensors. Fixed updating the date
//		when day changes. Length 7145 words.
// v0.4 8.10.2017
//		Changed ir_machine polling interval from 100s to 64s. This allows using any
//		of the timers to generate the required polling cycle. Changed timer interrupts
//		from timer 2 to timer 0 and 64 s. Frees timer 2 for driving buzzer.
//		Added hour beep function with possibility to set to short, long or off.
//		Better joystick operation with debounced press and release. Length 7457 words.
// v0.5 6.12.2017
//		Added support for RC5, Samsung and JVC IR protocols. Protocol is recognized 
//		automatically by the ir_machine() function. The identify menu now also shows
//		these codings. Length 7634 words.
// v0.6 23.12.2017
//		Added support for SIRC and NEC42 IR protocols. Rewrote RC5 decoder to sync
//		on every bit. Works now more reliably with different RC5 remotes. Added
//		hrbeep saving. Fixed displaying negative temperature values. Changed brightness
//		to show milliamperes instead of DAC values. Length 8503 words.
// v0.7 26.12.2017
//		Added support for time zones and automatic / manual daylight saving time.
//		Changed time registers to user struct_tm from time.h. Two structs, RTC uses always
//		utc_time. When performing RTC read, conversion from utc_time to local_time
//		is done, taking into account time zone and daylight saving time.
//		New include file timezone_converter.c has functions to deal with time zones
//		and daylight saving time. Added functions to pcf2129.c which work with struct_tm.
//		RTC is now read at 1 hour period instead of old 1 day period. This allows
//		proper way of checking daylight saving time. This approach needed the set time
//		and set date menus to update the RTC after each adjustment to prevent possible
//		time jump if minutes roll over to 0, causing RTC read of the 'old' time and date.
//		This causes some blinking in dimmed digits because the RTC read takes some 7 ms.
//		
//		9141 words when gps preliminary added, time in separate variables
//		9285 words when time changed to use struct_tm
//		9404 words when get_time_and_date changed to use struct_tm
//		9590 words when set_time_and_date changed to use struct_tm
//		11823 words when timezone and daylight saving time are in use
//		14007 words when GPS decoding active
//		13020 words with new GPRMC_decode_tm function which decodes time directly to struct_tm
//		bosch bme280 driver doesn't fit to program memory anymore. A total of about 17500 words
//		of program memory would be needed
//		12832 words with '#opt compress' optimization level defined and with bme280 driver,
//		support for showing humidity. Added decimal point location parameter to tlc5926
//		show number -function. Added support for detection of RTC and GPS at boot, and working
//		accordingly. Should work with any combination of these assembled, to be tested!
// v0.8 5.1.2018
//		Weird behavior with small unrelated changes in code cause e.g. temperature
//		displays to crash the FW. Updated compiler to 5.076. Program memory usage went
//		up from 12800 words to 15000 words, but the weird bugs disappeared. Now also
//		the bme280 seems to work properly in 'normal' mode, each bme280_get_humidity() is
//		returning fresh values without the need to first call bme280_set_mode() each time.
//		The #separate is needed again in bme280.c in order for the get_humidity function
//		to fit. Added support for displaying also bme280 temperature and pressure. bme280
//		is recognized at startup. The menu items to display bme280 sensor data are hidden
//		if the sensor is not found. Length 14757 words.
// v0.9 3.2.2018
//		Added option to show_number function to show temperature in same format as
//		show_temperature function. However using show_number function instead of
//		show_temperature functions consumes more program memory. Using new option
//		only on displaying BME temperatures of < -9.9C. Reformatted menu_machine()
//		function to use adjust_with_limits() function to scroll menu items. This
//		effectively avoids repeating event readout and state selection code.
//		The implementation required some reorganization of states and also usage
//		of some redirecting states.
//		14427 words after above changes
//		14267 words after changing tlc5925_show_string() to use 2-dimensional ascii array
//					and show_number and temperature functions to use 2-dimensional segment array
//		14639 words after changing tlc5925 to use more complete, 64 char ascii table
//		14644 words before changing tlc5925 to use display buffer for writes
//		14474 words just the unused 6-byte buffer added
//		14506 words when using tlc5925_send_buffer on show_temperature and show_number functions
//		14440 words when tlc5925_show_time() using get_digits() and tlc5925_send_buffer()
//		14408 words when tlc5925_show_date() using get_digits() and tlc5925_send_buffer()
//		14377 words when tlc5925_show_hex() using get_digits() and tlc5925_send_buffer()
//		14290 words when tlc5925_show_time_with_dimmer() using get_digits() and tlc5925_send_buffer()
//		14069 words when tlc5925_show_date_with_dimmer() using get_digits() and tlc5925_send_buffer()
//		now also fits to flash without #opt compress! 15485 words.
//		14235 words when bme_temperature shown on either in or out temperature if either ntc is not present
//		15522 words w/o opt compress.
//		12532 words when in and out temperatures use tlc5925_show_number function instead of show_temperature
//		15328 words w/o #opt compress.
//		14069 words when fixed seconds jumping with both RTC and GPS present. Required checking hour beep on
//		both RTC ISR and GPS_decode functions. using separate hourBeep() funtion for checking
//		If hourBeep() from GPS_decode commented out, memory usage 12514(!).
//		15338 words w/o #opt compress
//		12436 words when rewrote leading zero blanking and minus sign add routines to tlc5925_show_number funtion
//		14940 words w/o #opt compress
// v0.91 30.7.2018
//		12669 words when added 'toggle all' menu which toggles time, date, temperatures, pressure and humidity
//		does not fit w/o #opt compress
//		12604 words when using functions to show in and out temperatures
//		15200 words w/o #opt compress
//		14562 words when implemented ir send routine and added master / slave sync functionality via IR messages
//		15778 words w/o #opt compress
//		14361 words when all settings including menu_state are saved to eeprom and not to RTC alarm regs anymore
//		15865 words w/o #opt compress
//		14447 words when added 'region' setting with EU and US modes. Affects date and time formatting
//		does not fit w/o #opt compress
//		15969 words when added dewpoint calculation. NTC readout had to be changed from lookup table to
//		calculation to be able to fit. Region now affects also temperature scales (C or F).
//		15969 before changing to use parameter struct
//		14959 when changed parameters to use struct and menu and set menus to use string arrays for menu texts
//		15899 when seprated region setting into timeformat, dateformat and temperature unit settings
//		15968 when added day of week display
//		16119 when added alarm set menus but alarm not making sound yet
//		16019 when making sound and added time modes w/o seconds
//		15874 when added pending_b byte and set_brightness using that
//		15911 when changed alarm to have different tone than hour beep
//		16158 when reverted to using precalculated temp8 in menuMachine when adjusting parameters
//			  if it was calculated inside array brackets, caused weird display ghosting/blinking(?) #opt compress bug?
//		16159 when rerganized the loop a bit to prevent funny dewpoint if powered in dewpoint state
//		      bme280 sensor still is slightly off at first reading
//		16162 when initializing both IR_PACKET structs globally
//		16198 when changed bme280 init settings and added a delay to prevent errorneous bme280 reading at startup
//		15927 when moved daylight_saving_on calculation to separate function, which is now also called from decode_gps
//			  also fixed sending master clock time if timebase comes from GPS and slave mode with GPS
//		16223 when added support for Kaseikyo and Matsushita IR protocols. Also changed all timing values in
//			  ir_machine.c to parameters defined in ir_defines.h, which are set depending on interrupt frequency.
//			  This allows easier adaptation of the ir receiving function to different projects (and interrupt
//			  frequencies)
// v0.92 12.9.2018
//		16239 words when improved remote identify to also show unknown protocols and corrupted messages as 'Error'
//			  Remote identify also returns to 'none' after 4 seconds if nothing is received.
//		16232 words when added a separate wait_for_idle state to ir_machine to separate from error. Not saving state
//			  in setup menu anymore, only to clock, date, temp. etc. display states.
//		16049 words when added new setup menu item 'bAud' which selects between 9600 and 115200 rates for GPS
//			  receiver. 9600 is needed for U-Blox CAM-M8Q used in v4.1 board, 115200 is for Fastrax UC530M used
//			  in v4.0 board. Modified validity checking in gprmc_decoder to work with both GPS modules.
//		16023 words when fixed slave_receive() function which was probably broken when ir_machine() return values
//			  for NEC protocol were changed to include both high and low bytes in command. Implemented localize()
//			  function to save some program memory words. 

#opt compress
#define FW_VERSION 92
#define OPAMP_NI_TO_DAC1 0x04	// Needed because error in CCS include file

#define RTC_CLK PIN_A6
#define TLC5925_LE PIN_C0

#define IRin PIN_A5
#define IRED PIN_A7

#define REXT 1000	// TLC5925 current set resistor value in ohms

#define NTC_CHANNEL_IN 11
#define NTC_CHANNEL_OUT 13
// #define NTC_LOOKUP_TABLE

#define WEATHER_READOUT_DELAY 512	// Number of times to run the loop before reading
									// temperature/pressure/humidity values. 2048 = ca. 6 sec

#include <16F1788.h>
#device ADC=12;
#DEVICE PASS_STRINGS=IN_RAM // allows passing constant char arrays to tlc5925.c function
							// copies string from rom to ram when needed
#device WRITE_EEPROM=NOINT	// allows interrupt to occur while writing to EEPROM

#include <stdlib.h>
#include <stdio.h>
#define TIME_T_USES_2010
#include <time.h>
#include <math.h>	// for dewpoint calculation
#include <internal_eeprom.c>	// for reading / writing int16
#fuses INTRC_IO,NOWDT,NOFCMEN,NOIESO,BROWNOUT,BORV25,NOPROTECT,NOLVP,NODEBUG,PUT,NOMCLR,NOCLKOUT,NOVCAP
#use delay(int=16MHz, clock=32MHz)
#use FAST_IO(a)
#use FAST_IO(b)
#use FAST_IO(c)
#use FAST_IO(e)
#use spi(FORCE_HW,BAUD=1000000, BITS=8, SAMPLE_RISE)
#use i2c(master,sda=PIN_B3,scl=PIN_B0,fast=50000,stream=softai2c)
#define STREAM_I2C_BME280 softai2c
#use pwm(CCP1, OUTPUT=PIN_C2, TIMER=2, PWM_OFF)
// #use rs232(baud=115200,xmit=PIN_C6,rcv=PIN_C7,ERRORS)	// Fastrax UC530M
// #use rs232(baud=9600,xmit=PIN_C6,rcv=PIN_C7,ERRORS)		// U-blox CAM-M8Q
#use rs232(xmit=PIN_C6,rcv=PIN_C7,ERRORS, NOINIT)			// GPS UART port
#include "pcf2129.c"
#include "tlc5925.c"
#include "ir_machine.c"
#include "ir_send.c"
#include "ntc_temperature.c"
#include "gprmc_decoder.c"
#include "timezone_converter.c"
#include "bme280.c"

// EEPROM addresses for other settings than those defined in parameters[]
#define REMOTE_PROTOCOL_ADDR 16
#define REMOTE_ADDRESS_ADDR 18
#define REMOTE_UP_ADDR 20
#define REMOTE_DOWN_ADDR 22
#define REMOTE_LEFT_ADDR 24
#define REMOTE_RIGHT_ADDR 26
#define REMOTE_BUTTON_ADDR 28
#define STATE_ADDR 30

typedef enum {
        ST_MENU_SET = 0,
        ST_MENU_TIME,
        ST_MENU_DATE,
        ST_MENU_DAY,
        ST_MENU_IN_TEMPERATURE,
        ST_MENU_OUT_TEMPERATURE,
        ST_MENU_TOGGLE,
        ST_MENU_TOGGLE_ALL,
        ST_MENU_HUMIDITY,
        ST_MENU_DEWPOINT,
        ST_MENU_PRESSURE,
        ST_MENU_SET_REDIRECT,
        ST_TIME,
        ST_DATE,
        ST_DAY,
        ST_IN_TEMPERATURE,
        ST_OUT_TEMPERATURE,
        ST_TOGGLE,
        ST_TOGGLE_ALL,
        ST_HUMIDITY,
        ST_DEWPOINT,
        ST_PRESSURE,
        ST_MENU_SET_TIME,
        ST_MENU_SET_DATE,
        ST_MENU_SET_ALARM,
        ST_MENU_SET_ZONE,
        ST_MENU_SET_TIMEFORMAT,
        ST_MENU_SET_DATEFORMAT,
        ST_MENU_SET_TEMPUNIT,
        ST_MENU_SET_SAVING,
        ST_MENU_SET_BRIGHTNESS,
        ST_MENU_SET_HRBEEP,
        ST_MENU_SET_SYNC,
        ST_MENU_SET_BAUD,
        ST_MENU_SET_REMOTE,
        ST_MENU_SET_VERSION,
        ST_MENU_SET_RETURN,
        ST_SET_HOURS,
        ST_SET_DAY,
        ST_SET_ALARM,
        ST_SET_ZONE,
        ST_SET_TIMEFORMAT,
        ST_SET_DATEFORMAT,
        ST_SET_TEMPUNIT,
        ST_SET_SAVING,
        ST_SET_BRIGHTNESS,
        ST_SET_HRBEEP,
        ST_SET_SYNC,
        ST_SET_BAUD,
        ST_MENU_REMOTE_IDENTIFY_REDIRECT,
        ST_VERSION,
        ST_MENU_SET_RETURN_REDIRECT,
        ST_SET_MINUTES,
        ST_SET_SECONDS,
        ST_SET_MONTH,
        ST_SET_YEAR,
        ST_SET_ALHOURS,
        ST_SET_ALMINUTES,
        ST_MENU_REMOTE_IDENTIFY,
        ST_MENU_REMOTE_SHOW_CODE,
        ST_MENU_REMOTE_LEARN,
        ST_MENU_REMOTE_RETURN,
        ST_REMOTE_IDENTIFY,
        ST_REMOTE_SHOW_CODE,
        ST_LEARN_REMOTE_UP,
        ST_MENU_REMOTE_RETURN_REDIRECT,
        ST_LEARN_REMOTE_DOWN,
        ST_LEARN_REMOTE_LEFT,
        ST_LEARN_REMOTE_RIGHT,
        ST_LEARN_REMOTE_BUTTON,
        ST_UNDEFINED
} MENU_STATES;
MENU_STATES menu_state = ST_TIME;

typedef enum {
        EVT_NONE = 0,
        EVT_UP,
        EVT_DOWN,
        EVT_RIGHT,
        EVT_LEFT,
        EVT_BUTTON
} EVENTS;
EVENTS event = EVT_NONE;


int16 buzzerDuration=0;
int8 dlight_saving_on=0;
unsigned int8 temp8;
unsigned int8 displayDuration;

char const alarm_strings[3][7] = {
	"   off",
	"  once",
	"repeat"
};

char const timeformat_strings[5][7] = {
	"24mmss",
	"12mmss",
	"24mm  ",
	"12mm A",
	"24blan"
};

char const dateformat_strings[8][7] = {
	"ddmmyy",
	"mmddyy",
	"yymmdd",
	"yyddmm",
	" d myy",
	" m dyy",
	"yy m d",
	"yy d m"
};

char const tempunit_strings[2][3] = {
	" c",
	" f"
};

char const dlight_strings[3][5] = {
	" off",
	"  on",
	"auto"
};

char const hrbeep_strings[3][6] = {
	" none",
	"short",
	" long"
};

char const sync_strings[3][7] = {
	"   off",
	"master",
	" slave"
};

char const baud_strings[2][7] = {
	"  9600",
	"115200"
};

char const menu_strings[11][7] = {
	" setup",
	"  hour",
	"  date",
	"   day",
	"intemp",
	"outemp",
	"toggle",
	"togall",
	"humidi",
	"dewpnt",
	"pressu"
};

char const menu_set_strings[15][7] = {
	"sethou",
	"setdat",
	"set al",
	"t-zone",
	"hour f",
	"date f",
	"temp f",
	"daylig",
	"bright",
	"hrbeep",
	"  sync",
	"  baud",
	"remote",
	"versio",
	"return"
};

char const day_strings[7][7] = {
	"sunday",
	"monday",
	"tuesda",
	"wednes",
	"thursd",
	"friday",
	"saturd"
};

// Constant array which has sane / safe limits to parameters
// There are also 'factory default' values to use when run for the first time
// These are also used if any of the parameters stored in eeprom are outside
// of valid range. The order here must be same as the actual parameters
// in parameters[] variable, also the order must be same as the menu items
// for the parameters.
unsigned int8 const parameter_limits[12][3] = {
	//  Low limit	High limit	Default
		0,			23,			7,			// Alarm hours
		0,			59,			0,			// Alarm minutes
		0,			2,			0,			// Alarm (off, once, repeat)
		0,			24,			12,			// Timezone (offset 12 hours)
		0,			4,			0,			// Time format (24 hour, 12 hour, 24 hour w/o sec, 12 hour w/o sec, 24 hour with leading zero blanking)
		0,			7,			0,			// Date format (dd.mm.yy d.m.yy mm.dd.yy m.d.yy yy.mm.dd yy.m.d yy.dd.mm yy.d.m)
		0,			1,			0,			// Temperature unit (C or F)
		0,			2,			0,			// Daylight saving mode
		3,			30,			6,			// Display brightness, 28 current settings from 1.5 mA to 15 mA
		0,			2,			0,			// Hourbeep none, short, long
		0,			2,			0,			// Sync mode off, master, slave
		0,			1,			0			// GPS baud rate, 9600 or 115200 baud
};

unsigned int8 parameters[12];
#define P_ALHOURS 0
#define P_ALMINUTES 1
#define P_ALARM 2
#define P_TIMEZONE 3
#define P_TIMEFORMAT 4
#define P_DATEFORMAT 5
#define P_TEMPUNIT 6
#define P_DLIGHT_SAVING 7
#define P_BRIGHTNESS 8
#define P_HRBEEP 9
#define P_SYNC 10
#define P_BAUD 11

struct_tm local_time = {0,0,0,1,JANUARY,118,MONDAY,0};
struct_tm utc_time = {0,0,0,1,JANUARY,118,MONDAY,0};
struct_tm received_time = {0,0,0,1,JANUARY,118,MONDAY,0};

unsigned int8 joyshadow, joytimer,releasetimer;

int1 ir_received = FALSE;

// GPS stuff:
char GPSData[128];
unsigned int8 GPSDataPtr=0;
char rx_char;
int8 GPSDataReady = FALSE;
int8 GPS_status = 2;	// 0=no fix, 1=fix, 2=no GPS module found

int8 RTC_init_status;	// 0=OK, 1=osc stopped, 2=no RTC found
int8 bme_ok;			// 0=not initialized / found, 1=OK

int8 alarm_on=0;

//int8 ntc_ok=0;
//#bit IN_NTC_OK = ntc_ok.0
//#bit OUT_NTC_OK = ntc_ok.1

int1 IN_NTC_OK=0;
int1 OUT_NTC_OK=0;
int1 update_local=0;

int8 pending=0;
#bit PENDING_RTC_SAVE = pending.0
#bit PENDING_RTC_READ = pending.1
#bit PENDING_STATE_SAVE = pending.2
#bit PENDING_IR_DECODE = pending.3
#bit PENDING_REMOTE_SAVE = pending.4
#bit PENDING_MASTER_SEND = pending.5
#bit PENDING_BEEP_CHECK = pending.6
#bit PENDING_SETTINGS_SAVE = pending.7
int8 pending_b=0;
#bit PENDING_BRIGHTNESS_SET = pending_b.0
#bit PENDING_BAUD_RATE_SET = pending_b.1

#byte IOCAF = 0x0393	// interrupt on change flag refister for PORTA
#bit RTC_IOC_FLAG = IOCAF.6

IR_PACKET outPacket = {NEC_PROTOCOL,0x0000,0x0000,0x00};
IR_PACKET receivedPacket = {UNKNOWN_PROTOCOL,0x0000,0x0000,0x00};

int8 remoteProtocol;
int16 remoteAddress;
int16 remoteCommandUp;
int16 remoteCommandDown;
int16 remoteCommandLeft;
int16 remoteCommandRight;
int16 remoteCommandButton;

int1 decode_allowed=1;
int1 slave_sec_sync=0;
int16 slave_second_counter=0;

signed int32 bme280_temp=0;
unsigned int32 bme280_pressure=0;
unsigned int32 bme280_humidity=0;

// constants for accurate dew point calculation (Arden Buck equation):
// const float a = 6.1121;
const float b = 18.678;
const float c = 257.14;
const float d = 234.5;
float ym;
int16 dp;

#separate void read_jstk(void);
#separate void code_to_button(void);
void menuMachine();
void displayMachine();
#separate void adjust_with_limits(signed int8* adjust, signed int8 low_limit, signed int8 high_limit);
#separate void adjust_with_limits(int16* adjust, int16 low_limit, int16 high_limit);
#separate void save_remote();
void load_remote();
void handle_pending_tasks();
void handleBuzzer();
void decodeGPS();
void my_write_eeprom(int8 address, int8 value);
int8 my_read_eeprom(int8 address);
void show_bme_temperature();
void show_in_temperature();
void show_out_temperature();
#separate void hourBeep();
void sendMaster();
void slave_receive(void);
void calculate_dewpoint();
#separate void check_alarm();
void handle_pending_b_tasks();
void update_dlight();
void localize();

#int_ioc
void rtcclkisr(void) {
	if (GPS_status!=1 || parameters[P_SYNC]==2) {	// only increment time based on RTC interrupts if no GPS found or no GPS fix or slave mode
		if (parameters[P_SYNC]==1)
			PENDING_MASTER_SEND = 1;
		if (local_time.tm_sec<59)
			local_time.tm_sec++;
		else {
			local_time.tm_sec=0;
			if (local_time.tm_min<59)
				local_time.tm_min++;
			else {
				local_time.tm_min=0;
				PENDING_RTC_READ=1;			// fetch new time from RTC each hour. this updates also dst when necessary
			}	
		}
		PENDING_BEEP_CHECK = 1;
	}	
}

#int_timer0
void timer0isr(void) {
	// output_high(IRED);	// for measuring interrupt length. Typically around 20 s when decoding and comparing remote codes. Typically 25 s if slave mode. Some 10 s more if complete code received.

	ir_machine();
	if (!ir_received)
		ir_received = getIrPacket(&receivedPacket);

	
	if (ir_received) {
		if (parameters[P_SYNC] == 2)
			PENDING_IR_DECODE = 1;
		if (decode_allowed)
			code_to_button();
	}	
		
	read_ntc_sensors();

	read_jstk();
	
	if (alarm_on) {
		if (event != EVT_NONE) {
			alarm_on = 0;
			if (parameters[P_ALARM] == 1) {		// alarm = once
				parameters[P_ALARM] = 0;
				PENDING_SETTINGS_SAVE = 1;
			}	
			event = EVT_NONE;
			buzzerDuration=0;
		}
		
		if (slave_second_counter==7812)
			buzzerDuration=4096;
	}		

	menuMachine();
	if (buzzerDuration)
		buzzerDuration--;
	
	if (slave_sec_sync) {
		slave_sec_sync=0;
		slave_second_counter=500;		// Takes some delays into account to make slave clock to change seconds at exact time
	}	

	slave_second_counter++;
	if (slave_second_counter==15625) {
		slave_second_counter=0;
		if (parameters[P_SYNC]==2)	// Slave clock
			bit_set(IOCAF,6);		// set i/o change flag, causes RTC interrupt routine to be fired
		if (displayDuration)
			displayDuration--;
	}		
	
	// output_low(IRED);
}	

#int_rda
void uartisr(void)		// uart interrupt for receiving GPS NMEA messages
{
   rx_char = getc();
   if (GPSDataReady)
   	return;
   switch (rx_char)
   {
      case '$':
         GPSDataPtr = 0;
      break;
     
      case '\n':
      	if (GPSData[3] == 'R')		// $GPRMC message
         GPSDataReady = TRUE;
      break;
   }
   
   GPSData[GPSDataPtr++ & 0x7F] = rx_char;
     
} 

void main()
{
	bme280_init_t bme280_config;
	// bme280_init_t bme280_config = {BM280_OSRS_16X,BM280_OSRS_16X,BM280_OSRS_16X,BM280_TIMING_1000,BM280_FILTER_OFF};

	unsigned int16 bme_sampling_delay=0;
	setup_timer_0(T0_DIV_2); // set timer0 to 64 s (prescaler 2)
	enable_interrupts(INT_TIMER0);
	enable_interrupts(INT_IOC_A6_L2H);	// interrupt for 1 Hz clock signal from RTC chip Changed from H2L to L2H, maybe fixes second jump bug (is the RTC updated at low-to-high -transition?
	enable_interrupts(INT_RDA);	// for receiving GPS meassages via uart
	enable_interrupts(GLOBAL);
	output_low(IRED);			// ensure IRED stays off if not master
	set_tris_a(0b01111111);		// all input except ired
	set_tris_b(0b11111111);		// tri-state all port b pins, scl and sda will be controlled by #use_i2c code
	set_tris_c(0b10010000);		// SDI and RX are inputs
	set_tris_e(0x11111000);		
	port_a_pullups(0b01111111);	// pullups for joystick, ir receiver and rtc_clk. not for ired
	port_b_pullups(0b11000000); // pullups for pgc and pgd lines
	port_c_pullups(0b10000000); // pullup for rx input incase no gps module installed
	port_e_pullups(0b00001000); // mclr pullup
		
	setup_adc_ports( sAN11|sAN13,VSS_VDD);	// Temperature sensors
	setup_vref(VREF_ON | VREF_ADC_2v048 | VREF_COMP_DAC_2v048);
	setup_opamp2(OPAMP_ENABLED | OPAMP_HIGH_GBWP_MODE | OPAMP_NI_TO_DAC1);
	setup_dac(DAC_VSS_FVR);

	setup_adc(ADC_CLOCK_DIV_32);	// Gives 1s conversion clock cycle time (fastest recommended)
	set_adc_channel(11);
	sums_ready=0;	// enables ADC conversions

	output_low(PIN_C2);	// ensure the off state of buzzer doesn't present a DC voltage over the buzzer

	delay_ms(250);
	setup_ccp1(CCP_OFF);
	output_low(PIN_C1);	// /OE low = display on
	
	RTC_init_status = pcf2129_init();
	
	for (int i=P_ALHOURS; i<=P_BAUD; i++) {			// Read parameters and hours from eeprom to parameters[] array
		parameters[i] = read_eeprom(i);
		if (parameters[i] < parameter_limits[i][0] || parameters[i] > parameter_limits[i][1])	// Check each parameter for validity
			parameters[i] = parameter_limits[i][2];	// Reset to default if out of range
	}	

	menu_state = read_eeprom(STATE_ADDR);
	if (menu_state >= ST_UNDEFINED)		// Some sanity checks for values stored to SRAM
		menu_state = ST_TIME;
		
	load_remote();
	
	PENDING_RTC_READ = 1;
	PENDING_BRIGHTNESS_SET = 1;
	PENDING_BAUD_RATE_SET = 1;

	bme280_config.humidityOversampling = BM280_OSRS_16X;		// with max 16x oversampling on all measurements,
	bme280_config.temperatureOversampling = BM280_OSRS_2X;		// one measurement cycle takes about 100 ms
	bme280_config.pressureOversampling = BM280_OSRS_4X;			// we use values which provide low enough noise and 
																// low duty cycle to minimize selfheat of sensor
	bme280_config.inactiveDuration = BM280_TIMING_1000;			// 1 second sleep during measurement cycles (maximum)
	bme280_config.iirFilter = BM280_FILTER_1;					// 1 means filter coefficient of 16

	bme280_init(bme280_config);						// initialize the sensor with config values set above
	bme_ok = bme280_set_mode(BM280_MODE_NORMAL);	// bme280_set_mode returns false if bme not initialized or found

	if (bme_ok)										// delay to ensure that first bme280 measurement cycle is finished
		delay_ms(100);								// otherwise first displayed value after powerup is way off

	// Infinite loop
	while(1)
	{
		if (pending)
			handle_pending_tasks();
		if (pending_b)
			handle_pending_b_tasks();

		if (parameters[P_SYNC] == 2)
			disable_interrupts(INT_IOC_A6_L2H);
		else
			enable_interrupts(INT_IOC_A6_L2H);

		displayMachine();

		if (menu_state <= ST_PRESSURE) {	// BME sensor is not read when in setup
			if (bme_sampling_delay==0) {
				bme280_get_humidity(&bme280_temp, &bme280_pressure, &bme280_humidity);
			}	
		}		

		decodeGPS();
		
		handleBuzzer();

		if (menu_state < ST_MENU_SET_TIME && sums_ready && bme_sampling_delay==0) {		// don't do excess calculations when in set menus to prevent blinking
//			lookup_32bit(&in_sum, &in_temp);
//			lookup_32bit(&out_sum, &out_temp);
			IN_NTC_OK = calculate_32bit(&in_sum, &in_temp, (unsigned int16)(10000), (unsigned int16)(3950));
			OUT_NTC_OK = calculate_32bit(&out_sum, &out_temp, (unsigned int16)(10000), (unsigned int16)(3950));
			sums_ready=0;
		}	

		bme_sampling_delay++;
		if (bme_sampling_delay == WEATHER_READOUT_DELAY)
			bme_sampling_delay = 0;

		if (menu_state == ST_DEWPOINT)
			calculate_dewpoint();


	}	// end while
}

void menuMachine()
{
	temp8 = menu_state;
	switch(menu_state)
	{
		case ST_MENU_SET:
		case ST_MENU_TIME:
		case ST_MENU_DATE:
		case ST_MENU_DAY:
		case ST_MENU_IN_TEMPERATURE:
		case ST_MENU_OUT_TEMPERATURE:
		case ST_MENU_TOGGLE:
		case ST_MENU_TOGGLE_ALL:
		case ST_MENU_HUMIDITY:
		case ST_MENU_DEWPOINT:
		case ST_MENU_PRESSURE:
			switch(event)
			{
				case EVT_BUTTON:
//					if (menu_state != ST_MENU_SET)		// Do not save setup menu as previous used state - this was messing the state machine up for some reason
					if (temp8 != ST_MENU_SET)		// Do not save setup menu as previous used state
						PENDING_STATE_SAVE=1;
					menu_state += ST_TIME - ST_MENU_TIME;
					break;
				case EVT_LEFT:
				case EVT_RIGHT:
					if (bme_ok)
						adjust_with_limits(&menu_state, ST_MENU_SET, ST_MENU_PRESSURE);
					else
						adjust_with_limits(&menu_state, ST_MENU_SET, ST_MENU_TOGGLE_ALL);
					break;
				default:
			}
			break;	
		case ST_MENU_SET_REDIRECT:
//			menu_state = ST_MENU_SET_TIME;
			menu_state = ST_MENU_SET_RETURN;
			break;
		case ST_TIME:
		case ST_DATE:
		case ST_DAY:
		case ST_IN_TEMPERATURE:
		case ST_OUT_TEMPERATURE:
		case ST_TOGGLE:
		case ST_TOGGLE_ALL:
		case ST_HUMIDITY:
		case ST_DEWPOINT:
		case ST_PRESSURE:
			switch(event)
			{
				case EVT_BUTTON:
					menu_state -= ST_TIME - ST_MENU_TIME;
					break;
			}
			break;	
		case ST_MENU_SET_TIME:
		case ST_MENU_SET_DATE:
		case ST_MENU_SET_ALARM:
		case ST_MENU_SET_ZONE:
		case ST_MENU_SET_TIMEFORMAT:
		case ST_MENU_SET_DATEFORMAT:
		case ST_MENU_SET_TEMPUNIT:
		case ST_MENU_SET_SAVING:
		case ST_MENU_SET_BRIGHTNESS:
		case ST_MENU_SET_HRBEEP:
		case ST_MENU_SET_SYNC:
		case ST_MENU_SET_BAUD:
		case ST_MENU_SET_REMOTE:
		case ST_MENU_SET_VERSION:
		case ST_MENU_SET_RETURN:
			switch(event)
			{
				case EVT_BUTTON:
					menu_state += ST_SET_HOURS - ST_MENU_SET_TIME;
					break;
				case EVT_LEFT:
				case EVT_RIGHT:
					adjust_with_limits(&menu_state, ST_MENU_SET_TIME, ST_MENU_SET_RETURN);
					break;
				default:
			}
			break;
		case ST_SET_HOURS:
			switch(event)
			{
				case EVT_UP:
				case EVT_DOWN:
					adjust_with_limits(&local_time.tm_hour, 0, 23);
					PENDING_RTC_SAVE = 1;		// Added 2.1.2018 at v0.7 to fix hours jumping if minutes roll over, causing RTC read of old time
					break;
				case EVT_LEFT:
					menu_state = ST_SET_SECONDS;
					break;
				case EVT_RIGHT:
					menu_state = ST_SET_MINUTES;
					break;
				case EVT_BUTTON:
					menu_state = ST_MENU_SET_TIME;
					break;
			}
			break;
		case ST_SET_MINUTES:
			switch(event)
			{
				case EVT_UP:
				case EVT_DOWN:
					adjust_with_limits(&local_time.tm_min, 0, 59);
					PENDING_RTC_SAVE = 1;		// Added 2.1.2018 at v0.7 to fix minutes jumping if set to 59 and seconds roll over, causing RTC read of old time
					break;
				case EVT_BUTTON:
					menu_state = ST_MENU_SET_TIME;
					break;
				case EVT_LEFT:
					menu_state = ST_SET_HOURS;
					break;
				case EVT_RIGHT:
					menu_state = ST_SET_SECONDS;
					break;
			}
			break;
		case ST_SET_SECONDS:
			switch(event)
			{
				case EVT_BUTTON:
					menu_state = ST_MENU_SET_TIME;
					break;
				case EVT_LEFT:
					menu_state = ST_SET_MINUTES;
					break;
				case EVT_RIGHT:
					menu_state = ST_SET_HOURS;
					break;
				case EVT_NONE:
					break;
				default:
				//case EVT_UP:
				//case EVT_DOWN:
					if (local_time.tm_sec>29) {
						if (local_time.tm_min<59)
							local_time.tm_min++;
						else
							local_time.tm_min=0;
					}
					local_time.tm_sec=0;	
					PENDING_RTC_SAVE = 1;		// Added 2.1.2018 at v0.7
					break;
			}
			break;
		case ST_SET_DAY:
			switch(event)
			{
				case EVT_BUTTON:
					menu_state = ST_MENU_SET_DATE;
					break;
				case EVT_LEFT:
					menu_state = ST_SET_YEAR;
					break;
				case EVT_RIGHT:
					menu_state = ST_SET_MONTH;
					break;
				case EVT_NONE:
					break;
				default:
					adjust_with_limits(&local_time.tm_mday, 1, 31);
					PENDING_RTC_SAVE = 1;		// Added 2.1.2018 at v0.7
					break;
			}
			break;
		case ST_SET_MONTH:
			switch(event)
			{
				case EVT_BUTTON:
					menu_state = ST_MENU_SET_DATE;
					break;
				case EVT_LEFT:
					menu_state = ST_SET_DAY;
					break;
				case EVT_RIGHT:
					menu_state = ST_SET_YEAR;
					break;
				case EVT_NONE:
					break;
				default:
					adjust_with_limits(&local_time.tm_mon, 0, 11);
					PENDING_RTC_SAVE = 1;		// Added 2.1.2018 at v0.7
					break;
			}
			break;
		case ST_SET_YEAR:
			switch(event)
			{
				case EVT_BUTTON:
					menu_state = ST_MENU_SET_DATE;
					break;
				case EVT_LEFT:
					menu_state = ST_SET_MONTH;
					break;
				case EVT_RIGHT:
					menu_state = ST_SET_DAY;
					break;
				case EVT_NONE:
					break;
				default:
					adjust_with_limits(&local_time.tm_year, 118, 199);
					PENDING_RTC_SAVE = 1;		// Added 2.1.2018 at v0.7
					break;
			}
			break;
		case ST_SET_ALHOURS:
			switch(event)
			{
				case EVT_UP:
				case EVT_DOWN:
					adjust_with_limits(&parameters[P_ALHOURS], 0, 23);
					break;
				case EVT_LEFT:
				case EVT_RIGHT:
					menu_state = ST_SET_ALMINUTES;
					break;
				case EVT_BUTTON:
					menu_state = ST_MENU_SET_ALARM;
					PENDING_SETTINGS_SAVE = 1;
					break;
			}
			break;
		case ST_SET_ALMINUTES:
			switch(event)
			{
				case EVT_UP:
				case EVT_DOWN:
					adjust_with_limits(&parameters[P_ALMINUTES], 0, 59);
					break;
				case EVT_LEFT:
				case EVT_RIGHT:
					menu_state = ST_SET_ALHOURS;
					break;
				case EVT_BUTTON:
					menu_state = ST_MENU_SET_ALARM;
					PENDING_SETTINGS_SAVE = 1;
					break;
			}
			break;
		case ST_SET_ALARM:
		case ST_SET_ZONE:
		case ST_SET_TIMEFORMAT:
		case ST_SET_DATEFORMAT:
		case ST_SET_TEMPUNIT:
		case ST_SET_SAVING:
		case ST_SET_BRIGHTNESS:
		case ST_SET_HRBEEP:
		case ST_SET_SYNC:
		case ST_SET_BAUD:
			temp8 = menu_state - ST_SET_ALARM+2;		// +2 because P_ALHOURS and P_ALMINUTES are adjusted elsewhere
			switch(event)
			{
				case EVT_BUTTON:
					if (menu_state == ST_SET_ALARM && parameters[P_ALARM])
						menu_state = ST_SET_ALHOURS;
					else
						menu_state -= ST_SET_ALARM - ST_MENU_SET_ALARM;
					PENDING_SETTINGS_SAVE = 1;
					PENDING_RTC_READ = 1;
					PENDING_BAUD_RATE_SET = 1;		// only needed for baud rate but should't harm
					break;
				case EVT_NONE:
					break;
				case EVT_UP:
				case EVT_DOWN:
					adjust_with_limits(&parameters[temp8], parameter_limits[temp8][0], parameter_limits[temp8][1]);
					//adjust_with_limits(&parameters[menu_state - ST_SET_ALARM+2], parameter_limits[menu_state - ST_SET_ALARM+2][0], parameter_limits[menu_state - ST_SET_ALARM+2][1]);
					PENDING_BRIGHTNESS_SET = 1;	// only needed for brighness set but doesn't harm others
					break;
				default:
			}
			break;
		case ST_MENU_REMOTE_IDENTIFY_REDIRECT:
//			menu_state = ST_MENU_REMOTE_IDENTIFY;
			menu_state = ST_MENU_REMOTE_RETURN;
			break;
		case ST_VERSION:
			switch(event)
			{
				case EVT_BUTTON:
					menu_state = ST_MENU_SET_VERSION;
					break;
			}
			break;
		case ST_MENU_SET_RETURN_REDIRECT:	// this pseudo state is just to allow constant
			menu_state = ST_MENU_SET;		// jumps from setup menu items
			break;
		case ST_MENU_REMOTE_IDENTIFY:
		case ST_MENU_REMOTE_SHOW_CODE:
		case ST_MENU_REMOTE_LEARN:
		case ST_MENU_REMOTE_RETURN:
			decode_allowed=1;
			switch(event)
			{
				case EVT_BUTTON:
					menu_state += ST_REMOTE_IDENTIFY - ST_MENU_REMOTE_IDENTIFY;
					break;
				case EVT_LEFT:
				case EVT_RIGHT:
					adjust_with_limits(&menu_state, ST_MENU_REMOTE_IDENTIFY, ST_MENU_REMOTE_RETURN); 
					break;
			}
			break;
		case ST_REMOTE_IDENTIFY:
			switch(event)
			{
				case EVT_BUTTON:
					menu_state = ST_MENU_REMOTE_IDENTIFY;
					break;
			}
			break;
		case ST_REMOTE_SHOW_CODE:
			switch(event)
			{
				case EVT_BUTTON:
					menu_state = ST_MENU_REMOTE_SHOW_CODE;
					break;
			}
			break;
		case ST_LEARN_REMOTE_UP:
			decode_allowed=0;
			if (event == EVT_BUTTON)
				menu_state = ST_MENU_REMOTE_LEARN;
			if (ir_received) {
				remoteProtocol = receivedPacket.protocol;
				remoteAddress = receivedPacket.address;
				remoteCommandUp = receivedPacket.command;
				ir_received = FALSE;
				menu_state = ST_LEARN_REMOTE_DOWN;
			}	
			break;
		case ST_MENU_REMOTE_RETURN_REDIRECT:
			menu_state = ST_MENU_SET_REMOTE;
			break;
		case ST_LEARN_REMOTE_DOWN:
			if (event == EVT_BUTTON)
				menu_state = ST_MENU_REMOTE_LEARN;
			if (ir_received) {
				remoteCommandDown = receivedPacket.command;
				ir_received = FALSE;
				menu_state = ST_LEARN_REMOTE_LEFT;
			}	
			break;
		case ST_LEARN_REMOTE_LEFT:
			if (event == EVT_BUTTON)
				menu_state = ST_MENU_REMOTE_LEARN;
			if (ir_received) {
				remoteCommandLeft = receivedPacket.command;
				ir_received = FALSE;
				menu_state = ST_LEARN_REMOTE_RIGHT;
			}	
			break;
		case ST_LEARN_REMOTE_RIGHT:
			if (event == EVT_BUTTON)
				menu_state = ST_MENU_REMOTE_LEARN;
			if (ir_received) {
				remoteCommandRight = receivedPacket.command;
				ir_received = FALSE;
				menu_state = ST_LEARN_REMOTE_BUTTON;
			}	
			break;
		case ST_LEARN_REMOTE_BUTTON:
			if (event == EVT_BUTTON)
				menu_state = ST_MENU_REMOTE_LEARN;
			if (ir_received) {
				remoteCommandButton = receivedPacket.command;
				ir_received = FALSE;
				menu_state = ST_MENU_REMOTE_LEARN;
				PENDING_REMOTE_SAVE = 1;
			}	
			break;
	} // end menu case
	event = EVT_NONE;
}

void displayMachine()
{
	switch(menu_state)
	{
		case ST_MENU_TIME:
		case ST_MENU_DATE:
		case ST_MENU_DAY:
		case ST_MENU_IN_TEMPERATURE:
		case ST_MENU_OUT_TEMPERATURE:
		case ST_MENU_TOGGLE:
		case ST_MENU_TOGGLE_ALL:
		case ST_MENU_HUMIDITY:
		case ST_MENU_DEWPOINT:
		case ST_MENU_PRESSURE:
		case ST_MENU_SET:
			tlc5925_show_string(menu_strings[menu_state]);
			break;
		case ST_TIME:
			tlc5925_show_time(local_time.tm_hour, local_time.tm_min, local_time.tm_sec, parameters[P_TIMEFORMAT]);
			break;
		case ST_DATE:
			tlc5925_show_date(local_time.tm_mday, local_time.tm_mon+1, (int8)(local_time.tm_year), parameters[P_DATEFORMAT]);
			break;
		case ST_DAY:
			tlc5925_show_string(day_strings[local_time.tm_wday]);
			break;
		case ST_IN_TEMPERATURE:
			show_in_temperature();
			break;
		case ST_OUT_TEMPERATURE:
			show_out_temperature();
			break;
		case ST_TOGGLE:
			if(local_time.tm_sec%12<9)
				tlc5925_show_time(local_time.tm_hour, local_time.tm_min, local_time.tm_sec, parameters[P_TIMEFORMAT]);
			else
				tlc5925_show_date(local_time.tm_mday, local_time.tm_mon+1, (int8)(local_time.tm_year), parameters[P_DATEFORMAT]);
			break;
		case ST_TOGGLE_ALL:
			if(local_time.tm_sec%12<8)
				tlc5925_show_time(local_time.tm_hour, local_time.tm_min, local_time.tm_sec, parameters[P_TIMEFORMAT]);
			else if (local_time.tm_sec<24 && bme_ok) {
				if (local_time.tm_sec<12)
					tlc5925_show_number(bme280_humidity*100/1024, 2|0x08);
				else
					tlc5925_show_number(bme280_pressure, 2);
			}	
			else if (local_time.tm_sec==8 || local_time.tm_sec==32)
				tlc5925_show_string("in");
			else if (local_time.tm_sec==20 || local_time.tm_sec==44)
				tlc5925_show_string("out");
			else if (local_time.tm_sec<12 || (local_time.tm_sec>32 && local_time.tm_sec<37))
				show_in_temperature();
			else if (local_time.tm_sec<49)
				show_out_temperature();
			else
				tlc5925_show_date(local_time.tm_mday, local_time.tm_mon+1, (int8)(local_time.tm_year), parameters[P_DATEFORMAT]);
			break;
		case ST_HUMIDITY:
			tlc5925_show_number(bme280_humidity*100/1024, 2|0x08);
			//tlc5925_show_number((signed int32)(bme280_humidity*100/1024), 2|0x08);
			break;
		case ST_DEWPOINT:
			if (parameters[P_TEMPUNIT])
				tlc5925_show_number(dp*9/5+3200,(2|0x80));
			else
				tlc5925_show_number(dp,(2|0x20));
			break;
		case ST_PRESSURE:
			tlc5925_show_number(bme280_pressure, 2);
			//tlc5925_show_number((signed int32)(bme280_pressure), 2);
			break;
		case ST_MENU_SET_TIME:
		case ST_MENU_SET_DATE:
		case ST_MENU_SET_ALARM:
		case ST_MENU_SET_ZONE:
		case ST_MENU_SET_TIMEFORMAT:
		case ST_MENU_SET_DATEFORMAT:
		case ST_MENU_SET_TEMPUNIT:
		case ST_MENU_SET_SAVING:
		case ST_MENU_SET_BRIGHTNESS:
		case ST_MENU_SET_HRBEEP:
		case ST_MENU_SET_SYNC:
		case ST_MENU_SET_BAUD:
		case ST_MENU_SET_REMOTE:
		case ST_MENU_SET_VERSION:
		case ST_MENU_SET_RETURN:
			tlc5925_show_string(menu_set_strings[menu_state-ST_MENU_SET_TIME]);
			break;
		case ST_SET_HOURS:
			tlc5925_show_time_with_dimmer(local_time.tm_hour, local_time.tm_min, local_time.tm_sec, 1);	
			break;
		case ST_SET_MINUTES:
			tlc5925_show_time_with_dimmer(local_time.tm_hour, local_time.tm_min, local_time.tm_sec, 2);	
			break;
		case ST_SET_SECONDS:
			tlc5925_show_time_with_dimmer(local_time.tm_hour, local_time.tm_min, local_time.tm_sec, 3);	
			break;
		case ST_SET_DAY:
			tlc5925_show_date_with_dimmer(local_time.tm_mday, local_time.tm_mon+1, (int8)(local_time.tm_year), 1);
			break;
		case ST_SET_MONTH:
			tlc5925_show_date_with_dimmer(local_time.tm_mday, local_time.tm_mon+1, (int8)(local_time.tm_year), 2);
			break;
		case ST_SET_YEAR:
			tlc5925_show_date_with_dimmer(local_time.tm_mday, local_time.tm_mon+1, (int8)(local_time.tm_year), 3);
			break;
		case ST_SET_ALARM:
			tlc5925_show_string(alarm_strings[parameters[P_ALARM]]);
			break;
		case ST_SET_ALHOURS:
			tlc5925_show_time_with_dimmer(parameters[P_ALHOURS], parameters[P_ALMINUTES], 0, 1);	
			break;
		case ST_SET_ALMINUTES:
			tlc5925_show_time_with_dimmer(parameters[P_ALHOURS], parameters[P_ALMINUTES], 0, 2);	
			break;
		case ST_SET_ZONE:
			tlc5925_show_number((signed int8)(parameters[P_TIMEZONE])-12,0x07);
			break;
		case ST_SET_TIMEFORMAT:
			tlc5925_show_string(timeformat_strings[parameters[P_TIMEFORMAT]]);
			break;
		case ST_SET_DATEFORMAT:
			tlc5925_show_string(dateformat_strings[parameters[P_DATEFORMAT]]);
			break;
		case ST_SET_TEMPUNIT:
			tlc5925_show_string(tempunit_strings[parameters[P_TEMPUNIT]]);
			break;
		case ST_SET_SAVING:
			tlc5925_show_string(dlight_strings[parameters[P_DLIGHT_SAVING]]);
			break;
		case ST_SET_BRIGHTNESS:
			tlc5925_show_number((parameters[P_BRIGHTNESS]<<2)+parameters[P_BRIGHTNESS], 0x07);	// much less code than *5
			break;
		case ST_SET_HRBEEP:
			tlc5925_show_string(hrbeep_strings[parameters[P_HRBEEP]]);
			break;
		case ST_SET_SYNC:
			tlc5925_show_string(sync_strings[parameters[P_SYNC]]);
			break;
		case ST_SET_BAUD:
			tlc5925_show_string(baud_strings[parameters[P_BAUD]]);
			break;
		case ST_MENU_REMOTE_IDENTIFY:
			tlc5925_show_string("identi");
			break;
		case ST_REMOTE_IDENTIFY:
			/*
			if (receivedPacket.protocol) {
				repeatcounter++;
				receivedPacket.protocol=0;
			}
			tlc5925_show_number(repeatcounter);
			*/
			if (receivedPacket.protocol != 254) {
				switch (receivedPacket.protocol)
				{
					case UNKNOWN_PROTOCOL:
						tlc5925_show_string("error");
						break;
					case SIRC_PROTOCOL:
	//					tlc5925_show_string("sirc");
						if (bits_received==20)
							tlc5925_show_string("sirc20");
						else if (bits_received==15)
							tlc5925_show_string("sirc15");
						else
							tlc5925_show_string("sirc12");
						break;
					case NEC_PROTOCOL:
						tlc5925_show_string("nec");
						break;
					case SAMSUNG_PROTOCOL:
						tlc5925_show_string("samsun");
						break;
					case MATSUSHITA_PROTOCOL:
						tlc5925_show_string("matsus");
						break;
					case KASEIKYO_PROTOCOL:
						tlc5925_show_string("japan");
						break;
					case RC5_PROTOCOL:
						tlc5925_show_string("rc-5");
						break;
					case JVC_PROTOCOL:
						tlc5925_show_string("jvc");
						break;
					case NEC42_PROTOCOL:
						tlc5925_show_string("nec42");
						break;
				}
				displayDuration = 4;
				receivedPacket.protocol = 254;
			}
			if (displayDuration == 0)
				tlc5925_show_string("none");	
			break;
		case ST_MENU_REMOTE_SHOW_CODE:
			tlc5925_show_string("shocod");
			break;
		case ST_REMOTE_SHOW_CODE:
			tlc5925_show_hex(receivedPacket.address>>8, receivedPacket.address, receivedPacket.command);
			break;
		case ST_MENU_REMOTE_LEARN:
			tlc5925_show_string("learn");
			break;
		case ST_LEARN_REMOTE_UP:
			tlc5925_show_string("up");
			break;
		case ST_LEARN_REMOTE_DOWN:
			tlc5925_show_string("down");
			break;
		case ST_LEARN_REMOTE_LEFT:
			tlc5925_show_string("left");
			break;
		case ST_LEARN_REMOTE_RIGHT:
			tlc5925_show_string("right");
			break;
		case ST_LEARN_REMOTE_BUTTON:
			tlc5925_show_string("button");
			break;
		case ST_MENU_REMOTE_RETURN:
			tlc5925_show_string("return");
			break;
		case ST_VERSION:
			//tlc5925_show_string(FW_VERSION);
			tlc5925_show_number(FW_VERSION, 2);
			break;
	} // end display case
}

void read_jstk(void) {
	joyshadow = input_a() & 0x1f;
	
	if (joyshadow!=0x1f || joytimer)	// 15 ms debounce on press
		joytimer++;
	else
		joytimer=0;
	
	if (joyshadow==0x1f && releasetimer)			// 15 ms debounce on release
		releasetimer--;
	else if (joyshadow!=0x1f && joytimer>1)
		releasetimer=255;
	
	if (joytimer == 1 && releasetimer == 0) {
		if (!(joyshadow &0x04))
			event=EVT_DOWN;
		if (!(joyshadow &0x10))
			event=EVT_UP;
		if (!(joyshadow &0x02))
			event=EVT_LEFT;
		if (!(joyshadow &0x08))
			event=EVT_RIGHT;
		if (!(joyshadow &0x01))
			event=EVT_BUTTON;
	}		
}

void code_to_button(void) {
	event = EVT_NONE;
	if (receivedPacket.address == remoteAddress && receivedPacket.protocol == remoteProtocol) {
		if (receivedPacket.command == remoteCommandUp)
			event = EVT_UP;
		else if (receivedPacket.command == remoteCommandDown)
			event = EVT_DOWN;
		else if (receivedPacket.command == remoteCommandButton)
			event = EVT_BUTTON;
		else if (receivedPacket.command == remoteCommandLeft)
			event = EVT_LEFT;
		else if (receivedPacket.command == remoteCommandRight)
			event = EVT_RIGHT;
	}
	ir_received=FALSE;
}

void adjust_with_limits(signed int8* adjust, signed int8 low_limit, signed int8 high_limit) {
//	if (bit_test(event, 0)) {	// leads to same compiled output as the following line, but is less portable
	if (event&0x01) {	// this checks for both EVT_UP and EVT_RIGHT, but event enum must be properly arranged!
//	if (event==EVT_UP||event==EVT_RIGHT) {
		if (*adjust<high_limit)
			(*adjust)++;
		else
			*adjust=low_limit;
		break;
	}
	else {	
//	else if (event==EVT_DOWN||event==EVT_LEFT) {	// this check not needed if other events already filtered in state machine
		if (*adjust>low_limit)
			(*adjust)--;
		else
			*adjust=high_limit;
		break;
	}
}

void adjust_with_limits(int16* adjust, int16 low_limit, int16 high_limit) {
	if (event==EVT_UP) {
		if (*adjust<high_limit)
			(*adjust)++;
		else
			*adjust=low_limit;
		break;
	}
	else if (event==EVT_DOWN) {	
		if (*adjust>low_limit)
			(*adjust)--;
		else
			*adjust=high_limit;
		break;
	}
}

void save_remote() {
	my_write_eeprom(REMOTE_PROTOCOL_ADDR, remoteProtocol);
	write_int16_eeprom(REMOTE_ADDRESS_ADDR, remoteAddress);
	write_int16_eeprom(REMOTE_UP_ADDR, remoteCommandUp);
	write_int16_eeprom(REMOTE_DOWN_ADDR, remoteCommandDown);
	write_int16_eeprom(REMOTE_LEFT_ADDR, remoteCommandLeft);
	write_int16_eeprom(REMOTE_RIGHT_ADDR, remoteCommandRight);
	write_int16_eeprom(REMOTE_BUTTON_ADDR, remoteCommandButton);
}

void load_remote() {
	remoteProtocol = read_eeprom(REMOTE_PROTOCOL_ADDR);
	remoteAddress = read_int16_eeprom(REMOTE_ADDRESS_ADDR);
	remoteCommandUp = read_int16_eeprom(REMOTE_UP_ADDR);
	remoteCommandDown = read_int16_eeprom(REMOTE_DOWN_ADDR);
	remoteCommandLeft = read_int16_eeprom(REMOTE_LEFT_ADDR);
	remoteCommandRight = read_int16_eeprom(REMOTE_RIGHT_ADDR);
	remoteCommandButton = read_int16_eeprom(REMOTE_BUTTON_ADDR);
}

void handle_pending_tasks() {
	if (PENDING_RTC_READ && parameters[P_SYNC]!=2 && RTC_init_status<2) {		// RTC_READ takes about 7 ms when localtime and daylight saving time implemented
//		pcf2129_get_time_and_date(utc_time.tm_sec, utc_time.tm_min, utc_time.tm_hour, utc_time.tm_mday, utc_time.tm_wday, utc_time.tm_mon, (int8)(utc_time.tm_year));
//		utc_time.tm_mon -= 1;
//		utc_time.tm_year += 100;
		pcf2129_get_struct_tm(&utc_time);	// get struct adjusts month and year according to struct_tm spec
		localize();
//		update_dlight();
//		utc_to_local(&utc_time, &local_time, (signed int8)(parameters[P_TIMEZONE])-12+dlight_saving_on);
	}	
	
	if (PENDING_RTC_SAVE) {
		if ((utc_time.tm_mon == 4 || utc_time.tm_mon == 6 || utc_time.tm_mon == 9 || utc_time.tm_mon == 11)&& utc_time.tm_mday > 30)
			utc_time.tm_mday = 30;
		if (utc_time.tm_year%4 == 0 && utc_time.tm_mon == 2 && utc_time.tm_mday>29)
			utc_time.tm_mday = 29;
		else if (utc_time.tm_mon == 2 && utc_time.tm_mday>28)
			utc_time.tm_mday = 28;
		utc_to_local(&local_time, &utc_time, -((signed int8)(parameters[P_TIMEZONE])-12+dlight_saving_on));	// use negation of timezone when converting local to utc
		pcf2129_set_time_and_date(utc_time.tm_sec, utc_time.tm_min, utc_time.tm_hour, utc_time.tm_mday, utc_time.tm_wday, utc_time.tm_mon+1, (int8)(utc_time.tm_year-100));	// this might produce less code since this function is internally used by pcf2129.c in init funcion
//		pcf2129_set_struct_tm(&utc_time); // alternative to line above, need to check which one produces more compact code
		
	}	
	

	if (PENDING_STATE_SAVE) {
		my_write_eeprom(STATE_ADDR, menu_state);
	}	
	
	if (PENDING_IR_DECODE) {
		slave_receive();
	}	
	
	if (PENDING_REMOTE_SAVE) {
		save_remote();
	}

	if (PENDING_MASTER_SEND) {
		sendMaster();
	}	

	if (PENDING_BEEP_CHECK) {
		hourBeep();
		check_alarm();
	}	

	if (PENDING_SETTINGS_SAVE) {
		for (int i=P_ALHOURS; i<=P_BAUD; i++) {
			my_write_eeprom(i, parameters[i]);
		}
	}	

	pending=0;
}		

void handleBuzzer() {
	if (buzzerDuration)
		setup_ccp1(CCP_PWM);
	else
		setup_ccp1(CCP_OFF);
}

void decodeGPS() {
	if (GPSDataReady) {
		if (GPS_status==2)		// First message received from GPS module, so it must be present!
			GPS_status=0;
		if (parameters[P_SYNC]!=2) {		// not slave clock
			GPS_status = GPRMC_decode_tm(GPSData, &received_time);
			if (GPS_status) {		// GPS data is valid (has a fix) 
				utc_time = received_time;
				if (utc_time.tm_min==59 && utc_time.tm_sec==0)	// If GPS has fix, save time&date to RTC once per hour
					PENDING_RTC_SAVE=1;
				localize();
				//update_dlight();
				//utc_to_local(&utc_time, &local_time, (signed int8)(parameters[P_TIMEZONE])-12+dlight_saving_on);
				PENDING_BEEP_CHECK = 1;	// cases GPS_status==0 and GPS_status==2 are handled on RTC ISR
				if (parameters[P_SYNC]==1)		// If GPS has fix and master mode, send time
					PENDING_MASTER_SEND = 1;
			}
		}		
		GPSDataReady = FALSE;
	}
}

void my_write_eeprom(int8 address, int8 value) {
    write_eeprom(address,value);
}

int8 my_read_eeprom(int8 address) {
    return read_eeprom(address);
}

void show_bme_temperature() {
	if (bme280_temp<-999)	// temperatures less than -9.99 must be shown with one decimal
		if (parameters[P_TEMPUNIT])
			tlc5925_show_number(bme280_temp*9/5+3200,(2|0x80));
		else
			tlc5925_show_number(bme280_temp,(2|0x20));
	else
		if (parameters[P_TEMPUNIT])
			tlc5925_show_number(bme280_temp*900/5+320000,(4|0x40));
		else
			tlc5925_show_number(bme280_temp*100,(4|0x10));
}

void show_in_temperature() {
	if (IN_NTC_OK)
		if (parameters[P_TEMPUNIT])
			tlc5925_show_number((signed int32)(in_temp*90/5+3200),(2|0x80));
		else
			tlc5925_show_number((signed int32)(in_temp*10),(2|0x20));
	else if (bme_ok)
		show_bme_temperature();
	else
		tlc5925_show_string("n/c");
}		

void show_out_temperature() {
	if (OUT_NTC_OK)
		if (parameters[P_TEMPUNIT])
			tlc5925_show_number((signed int32)(out_temp*90/5+3200),(2|0x80));
		else
			tlc5925_show_number((signed int32)(out_temp*10),(2|0x20));
	else if (IN_NTC_OK && bme_ok)
		show_bme_temperature();
	else
		tlc5925_show_string("n/c");
}

void hourBeep() {
	if (local_time.tm_sec>54 && local_time.tm_min==59 && parameters[P_HRBEEP]==2) { // beep mode = long
		setup_timer_2(T2_DIV_BY_64,125,1); // 1 kHz (8 MHz / 64 / 125)
		set_pwm1_duty(62);
		buzzerDuration=1562;
	}
	if (local_time.tm_sec==0 && local_time.tm_min==0 && parameters[P_HRBEEP]) {
		setup_timer_2(T2_DIV_BY_64,125,1); // 1 kHz (8 MHz / 64 / 125)
		set_pwm1_duty(62);
		buzzerDuration=7812;
	}	
}

void sendMaster() {
	
	switch(local_time.tm_sec) {
		case 0:
			outPacket.address = 0x7133;
			outPacket.command = utc_time.tm_hour;
			break;
		case 1:
			outPacket.address = 0x7132;
			outPacket.command = utc_time.tm_mday;
			break;
		case 2:
			outPacket.address = 0x7131;
			outPacket.command = utc_time.tm_mon;
			break;
		case 3:
			outPacket.address = 0x7130;		// year address
			outPacket.command = utc_time.tm_year;
			break;
		case 4:
			outPacket.address = 0x7134;
			outPacket.command = local_time.tm_min;
			break;
		break;
	}
	if (local_time.tm_sec < 5) {
		disable_interrupts(GLOBAL);
		ir_send(&outPacket);
		enable_interrupts(GLOBAL);
	}
}

void slave_receive(void) {
	unsigned int8 received_command;
	received_command = (unsigned int8)(receivedPacket.command);
	update_local = 0;
	if (receivedPacket.protocol == NEC_PROTOCOL && (unsigned int8)(receivedPacket.address>>8) == 0x71) {
		if ((unsigned int8)receivedPacket.address == 0x30) {
			utc_time.tm_year = received_command;
			update_local = 1;
		}	
		else if ((unsigned int8)receivedPacket.address == 0x31)
			utc_time.tm_mon = received_command;
		else if ((unsigned int8)receivedPacket.address == 0x32)
			utc_time.tm_mday = received_command;
		else if ((unsigned int8)receivedPacket.address == 0x33) {
			utc_time.tm_hour = received_command;
			update_local = 1;
		}	
		else if ((unsigned int8)receivedPacket.address == 0x34) {
			local_time.tm_min = received_command;
			local_time.tm_sec = 4;
			slave_sec_sync = 1;
		}	
		if (update_local) {
			update_dlight();
			utc_to_local(&utc_time, &received_time, (signed int8)(parameters[P_TIMEZONE])-12+dlight_saving_on);
			local_time.tm_year = received_time.tm_year;
			local_time.tm_mon = received_time.tm_mon;
			local_time.tm_mday = received_time.tm_mday;
			local_time.tm_wday = received_time.tm_wday;
			local_time.tm_hour = received_time.tm_hour;
		}	
	}
}

void calculate_dewpoint() {
     // calculate dew point with  Arden Buck equation:
     float t = bme280_temp/100.0;
     ym = log ( (bme280_humidity/102400.0)*exp((b-t/d)*(t/(c+t))));
     dp = 100*c*ym/(b-ym);
}

void check_alarm() {
	if (parameters[P_ALARM]) {
		if (local_time.tm_hour == parameters[P_ALHOURS] && local_time.tm_min == parameters[P_ALMINUTES] && (local_time.tm_sec == 0 || alarm_on == 1)) {
			setup_timer_2(T2_DIV_BY_64,31,1); // 4 kHz (8 MHz / 64 / 31)
			set_pwm1_duty(15);
			buzzerDuration=4096;
			alarm_on=1;
		}
	}		
}

void handle_pending_b_tasks() {
	if (PENDING_BRIGHTNESS_SET) {
		dac_write(151 - parameters[P_BRIGHTNESS]*REXT/288);
	}
	
	if (PENDING_BAUD_RATE_SET) {
		if (parameters[P_BAUD])
			setup_uart(115200);	// e.g. Fastrax UC530M
		else
			setup_uart(9600);		// e.g. U-blox CAM-M8Q
	}		
	pending_b = 0;
}	

void update_dlight() {
	if (parameters[P_DLIGHT_SAVING]<2)
		dlight_saving_on = parameters[P_DLIGHT_SAVING];
	else
		dlight_saving_on = eu_summertime(&utc_time);
}

void localize() {
	update_dlight();
	utc_to_local(&utc_time, &local_time, (signed int8)(parameters[P_TIMEZONE])-12+dlight_saving_on);
}	
